"""Streams defintions and abstract classes

Module contains abstract classes and implementations for
Stream patterns, such as:
    - Publisher or Emitter
    - Observer or Subscriber
    - Ventilator
    - Worker
    - Sink
"""

from . import mixin, zeromq

from .mixin import *
from .zeromq import *



"""motiv synchronization primitives

Module:
    Contains different types of internally used
    events
"""

from multiprocessing import Event as PEvent
from threading import Timer, Event as TEvent
from motiv.sync import mixin


class ProcessEvent(mixin.SystemEvent):

    """Event synchronization primitive.
    Wraps `multiprocessing.Event`
    """
    def __init__(self, *args, **kwargs):
        self.event = PEvent(*args, **kwargs)

    def set(self):
        """sets the event"""
        return self.event.set()

    def clear(self):
        """clears the event"""
        return self.event.clear()

    def wait(self, *args, **kwargs):
        """waits till event is set"""
        return self.event.wait(*args, **kwargs)

    def is_set(self):
        """checks if the event is set

        Returns:
            bool: True if set, False if cleared
        """
        return self.event.is_set()


class ThreadEvent(mixin.SystemEvent):

    """Event synchronization primitive.
    Wraps `threading.Event`
    """
    def __init__(self, *args, **kwargs):
        self.event = TEvent(*args, **kwargs)

    def set(self):
        """sets the event"""
        return self.event.set()

    def clear(self):
        """clears the event"""
        return self.event.clear()

    def wait(self, *args, **kwargs):
        """waits till event is set"""
        return self.event.wait(*args, **kwargs)

    def is_set(self):
        """checks if the event is set

        Returns:
            bool: True if set, False if cleared
        """
        return self.event.is_set()


class TimeoutEvent(ThreadEvent):

    """Timeout event that is automatically set
    after a specified timeout
    Args:
        timeout(int): timeout in millis

    """

    def __init__(self, timeout, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.timer = Timer(timeout * 1e-3, self.set)

    def start(self):
        """Starts the timer for the timeout"""
        return self.timer.start()


__all__ = [
        'ProcessEvent',
        'ThreadEvent',
        'TimeoutEvent'
        ]



"""motiv exceptions

"""


class ActorInitializationError(Exception):
    """Exception to denote failure to initialize an actor"""


class AlreadyConnected(Exception):
    """Exception that occurs connecting more than once"""


class NotConnected(Exception):
    """Exception to denote that a channel isn't connected"""


class ScalingError(Exception):
    """Exception to indicate an error encountered whiel scaling group"""


class DispatchError(Exception):
    """Exception to denote issue to dispatch, stop an actor"""



"""Streams interfaces and abstract classess

Module contains base classes for stream patterns that
must be derived by any stream implementation
"""

import abc


class SenderType(abc.ABC):
    """Sender stream type"""
    @abc.abstractmethod
    def send(self, payload, sync):
        """abstract send method"""


class ReceiverType(abc.ABC):
    """Receiver stream type"""
    @abc.abstractmethod
    def receive(self):
        """abstract receive method"""


class EmitterType(SenderType):
    """Emitter stream abstract class"""

    @abc.abstractmethod
    def connect(self):
        """abstract connect method"""

    def send(self, payload, sync=True):
        """sends a payload over underlying channel

        Args:
            payload: data to send over the channel
        """
        return self.channel_out.send(payload, sync)

    def close(self):
        """closes underlying channel"""
        return self.channel_out.close()

    @property
    @abc.abstractmethod
    def channel_out(self):
        """output channel"""


class SubscriberType(ReceiverType):
    """Subscriber stream abstract class"""

    @abc.abstractmethod
    def connect(self):
        """abstract connect method"""

    @abc.abstractmethod
    def subscribe(self, topic: int):
        """subscribe to a topic

        Args:
            topic (int): topic to subscribe to.
        """

    def receive(self):
        """block on the input channel for received data"""
        return self.channel_in.receive()

    def poll(self, *args, **kwargs):
        """polls input channel for received data"""
        return self.channel_in.poll(*args, **kwargs)

    def close(self):
        """closes stream's underlying channel"""
        return self.channel_in.close()

    @property
    @abc.abstractmethod
    def channel_in(self):
        """input channel"""


class VentilatorType(SenderType):
    """Ventilator stream abstract class"""

    @abc.abstractmethod
    def connect(self):
        """abstract connect method"""

    def send(self, body, sync=True):
        """sends a payload over underlying channel

        Args:
            payload: data to send over the channel
        """
        self.channel_out.send(body, sync)

    def close(self):
        """closes underlying channel"""
        return self.channel_out.close()

    @property
    @abc.abstractmethod
    def channel_out(self):
        """output channel"""


class WorkerType(ReceiverType):
    """Worker stream abstract class"""

    @abc.abstractmethod
    def connect(self):
        """abstract connect method"""

    def receive(self):
        """block on the input channel for received data"""
        return self.channel_in.receive()

    def poll(self, *args, **kwargs):
        """polls input channel for received data"""
        return self.channel_in.poll(*args, **kwargs)

    def close(self):
        """closes underlying channel"""
        return self.channel_in.close()

    @property
    @abc.abstractmethod
    def channel_in(self):
        """input channel"""

class PusherType(SenderType):
    """pushes data into a pulling stream (sink)"""

    @abc.abstractmethod
    def connect(self):
        """abstract connect method"""

    def send(self, body, sync=True):
        """sends a payload over underlying channel

        Args:
            payload: data to send over the channel
        """
        self.channel_out.send(body, sync)

    def close(self):
        """closes underlying channel"""
        return self.channel_out.close()

    @property
    @abc.abstractmethod
    def channel_out(self):
        """output channel"""


class SinkType(ReceiverType):
    """Sink stream abstract class"""

    @abc.abstractmethod
    def connect(self):
        """abstract connect method"""

    def receive(self):
        """block on the input channel for received data"""
        return self.channel_in.receive()

    def poll(self, *args, **kwargs):
        """polls input channel for received data"""
        return self.channel_in.poll(*args, **kwargs)

    def close(self):
        """closes underlying channel"""
        return self.channel_in.close()

    @property
    @abc.abstractmethod
    def channel_in(self):
        """input channel"""


class CompoundStreamType(SenderType, ReceiverType):
    """A type of two merged streams, ideally sender and receiver stream"""
    def send(self, payload, sync=True):
        """sends a payload over underlying channel

        Args:
            payload: data to send over the channel
        """
        self.stream_out.send(payload, sync)

    def receive(self):
        """block on the input channel for received data"""
        self.stream_in.receive()

    def poll(self, *args, **kwargs):
        """polls input channel for received data"""
        self.stream_in.poll(*args, **kwargs)

    @abc.abstractmethod
    def connect(self):
        """establish connection"""

    def close(self):
        """closes underlying streams"""
        self.stream_in.close()
        self.stream_out.close()


__all__ = [
        'EmitterType',
        'SubscriberType',
        'VentilatorType',
        'WorkerType',
        'PusherType',
        'SinkType',
        'CompoundStreamType',
        ]



"""Interfaces for serializable types"""

import abc


class Serializable(abc.ABC):
    """
    A De/Serializable interface
    """
    @abc.abstractmethod
    def serialize(self):
        """returns encoded object"""

    @classmethod
    @abc.abstractmethod
    def deserialize(cls, payload):
        """decodes an object into an instance"""



__all__ = [
        'Serializable'
        ]



"""Streams implementation using 0MQ

Todo:
    * check address format
"""
import zmq

from ensure import ensure_annotations, ensure
from motiv.streams import mixin
from motiv.channel import Channel, ChannelOut, ChannelIn


class Sender(mixin.SenderType):
    """ZMQ exclusive SenderType"""


class Receiver(mixin.ReceiverType):
    """ZMQ exclusive ReceiverType"""


class Emitter(mixin.EmitterType, Sender):
    """ZMQ Publishing stream

    Args:
        address(str): address to connect to (e.g. /tmp/socket).
    """
    def __init__(self, address: str):
        self.address = address
        self._cout = ChannelOut(zmq.PUB, address)

    def publish(self, topic, payload, sync=True):
        """Publishes data over a topic

        Args:
            topic(int): topic to broadcast payload over.
            payload: data to publish
        """
        _topic = bytes([topic])
        return self.send([_topic, payload], sync)

    def connect(self):
        """establish connection"""
        self.channel_out.bind()

    @property
    def channel_out(self):
        """output channel (readonly)"""
        return self._cout

class Subscriber(mixin.SubscriberType, Receiver):

    """ZMQ Subscriber stream

    Args:
        address(str): address to connect to (e.g. /tmp/socket).
    """

    @ensure_annotations
    def __init__(self, address: str):
        self.address = address
        self._cin = ChannelIn(zmq.SUB, address)

    def subscribe(self, topic: int):
        """Subscribes to a topic

        Args:
            topic(int): topic to subscribe to.
        """
        ensure(topic).is_an(int)
        _topic = bytes([topic])
        self.channel_in.sock_in.setsockopt(zmq.SUBSCRIBE, _topic)

    def connect(self):
        """establish connection"""
        self.channel_in.connect()

    @property
    def channel_in(self):
        """input channel (readonly)"""
        return self._cin


class Ventilator(mixin.VentilatorType, Sender):

    """ZMQ Ventilator

    Args:
        address(str): address to connect to (e.g. /tmp/socket).
    """
    @ensure_annotations
    def __init__(self, address: str):
        self.address = address
        self._cout = ChannelOut(zmq.PUSH, address)

    def connect(self):
        """establish connection"""
        return self.channel_out.bind()

    @property
    def channel_out(self):
        """output channel (readonly)"""
        return self._cout


class Worker(mixin.WorkerType, Receiver):
    """ZMQ Worker

    Args:
        address(str): address to connect to (e.g. /tmp/socket).
    """

    @ensure_annotations
    def __init__(self, address: str):
        self.address = address
        self._cin = ChannelIn(zmq.PULL, address)

    def connect(self):
        """establish connection"""
        return self.channel_in.connect()

    @property
    def channel_in(self):
        """input channel (readonly)"""
        return self._cin


class Pusher(mixin.PusherType, Sender):
    """ZMQ Pusher

    Args:
        address(str): address to connect to (e.g. /tmp/socket)
    """
    @ensure_annotations
    def __init__(self, address: str):
        self.address = address
        self._cout = ChannelOut(zmq.PUSH, address)

    def connect(self):
        """establish connection"""
        return self.channel_out.connect()

    @property
    def channel_out(self):
        """output channel (readonly)"""
        return self._cout


class Sink(mixin.SinkType, Receiver):
    """ZMQ Sink

    Args:
        address(str): address to connect to (e.g. /tmp/socket).
    """

    @ensure_annotations
    def __init__(self, address: str):
        self.address = address
        self._cin = ChannelIn(zmq.PULL, address)

    def connect(self):
        """establish connection"""
        return self.channel_in.bind()

    @property
    def channel_in(self):
        """input channel (readonly)"""
        return self._cin


class CompoundStream(mixin.CompoundStreamType, Sender, Receiver):
    """A stream consisting of two ZMQ Sreams

    Args:
        stream_in: receiver stream.
        stream_out: sender stream.

    Note:
        stream_in and stream_out can be identical.
    """

    @ensure_annotations
    def __init__(self, stream_in: Receiver, stream_out: Sender):
        self._stream_in = stream_in
        self._stream_out = stream_out
        self.channel = Channel(stream_in.channel_in, stream_out.channel_out)

    def run(self):
        """Starts a proxy over input and output streams"""
        self.channel.proxy()

    def connect(self):
        self.stream_in.connect()
        self.stream_out.connect()

    @property
    def channel_in(self):
        """input channel"""
        return self.channel

    @property
    def channel_out(self):
        """output channel"""
        return self.channel

    @property
    def stream_in(self):
        """output stream"""
        return self._stream_in

    @property
    def stream_out(self):
        """input stream"""
        return self._stream_out


__all__ = [
        'Emitter',
        'Subscriber',
        'Ventilator',
        'Worker',
        'Pusher',
        'Sink',
        'CompoundStream',
        'Receiver',
        'Sender'
        ]



"""
Module:
    motiv.version

Description:
    contains constants defining the package version

Note:
    - In case of a bugfix/hotfix increment BUILD_NUMBER
    - In case of adding a feature that modifies the package,
        but does not break interfaces, increment VERSION_MINOR
    - In case of introducing a change that breaks
        backward-comptability, increment VERSION_MAJOR
"""

def get_version():
    VERSION_MAJOR = 0
    VERSION_MINOR = 1
    BUILD_NUMBER = 0

    VERSION = f"{VERSION_MAJOR}.{VERSION_MINOR}.{BUILD_NUMBER}"
    return VERSION

__version__ = get_version()
__all__ = ['__version__']



"""multiprocessing based actor and events"""

import abc
import signal

from multiprocessing import Process
from ensure import ensure_annotations, ensure

from motiv.sync import ProcessEvent
from motiv.exceptions import ActorInitializationError
from motiv.channel import zeromq as zchannels
from motiv.streams import zeromq as zstreams


class ExecutionContextBase(Process, abc.ABC):

    """Mixin containing default behavior of an actor loop.

    Args:
        name(str): actor name

    Attributes:
        name(str): actor name
    """
    def __init__(self, name: str):
        self.name = name
        self._halt = ProcessEvent()
        super().__init__(name=name)

    def initialize_context(self):
        """setups context environment

        this function should not be overloaded
        outside of the package.
        """

        def __signal_handler(_signum, _frame):
            self.stop()

        # Setup signal handling.
        signal.signal(signal.SIGABRT, __signal_handler)
        signal.signal(signal.SIGTERM, __signal_handler)
        signal.signal(signal.SIGCHLD, __signal_handler)
        signal.signal(signal.SIGINT, __signal_handler)
        signal.signal(signal.SIGILL, __signal_handler)

    @abc.abstractmethod
    def pre_start(self):
        """executes before starting an actor

        Note:
            this method runs in the actor child-process
            meaning streams, channels, sockets initialized
            in this function is owned by the actor process
        """

    @abc.abstractmethod
    def post_stop(self):
        """executes after the actor has stopped
        close streams, sockets, descriptors in this function
        """

    def stop(self):
        """halts the actor"""
        return self._halt.set()

    @property
    def runnable(self):
        """bool: True if actor hasn't been signaled to stop,False otherwise"""
        return not self._halt.is_set()


class ExecutionContext(ExecutionContextBase):
    """Actual actor definition.

    Args:
        name(str): actor name.

    Attributes:
        name: actor name.
        poller: stream poller.
    """

    def __init__(self, name):
        self.name = name
        super().__init__(self.name)
        self._stream_in = None
        self._stream_out = None
        self._stream = None
        self.poller = zchannels.Poller()

        # Properties
        self._poll_timeout = 5

    def receive(self, **kwargs):
        """polls input stream for data"""
        if not self.stream_in:
            raise ActorInitializationError("No in-stream set")
        return self.stream_in.poll(self.poller, self._halt, **kwargs)

    def send(self, payload, sync=True):
        """sends data over the output stream"""
        if not self.stream_out:
            raise ActorInitializationError("No out-stream set")
        return self.stream_out.send(payload, sync)

    def publish(self, topic, payload):
        """publishes data over the output stream"""
        ensure(self.stream_out).is_a(zstreams.Emitter)
        return self.stream_out.publish(topic, payload)

    @abc.abstractmethod
    def run(self):
        """process body"""

    @ensure_annotations
    def set_stream(self, stream):
        """assigns a given stream to out-stream, in-stream."""
        if isinstance(stream, zstreams.Sender) and \
                isinstance(stream, zstreams.Receiver):
            self.stream = stream
        elif isinstance(stream, zstreams.Sender):
            self.stream_out = stream
        elif isinstance(stream, zstreams.Receiver):
            self.stream_in = stream
        else:
            raise ValueError("Incompatible stream type")

    @property
    def stream_in(self):
        """actor input stream"""
        return self._stream_in

    @stream_in.setter
    @ensure_annotations
    def stream_in(self, value: zstreams.Receiver):
        if self.stream is not None:
            raise ValueError("stream_in is set from stream,"
                             " reset stream to overwrite")
        self._stream_in = value

    @property
    def stream_out(self):
        """actor output stream"""
        return self._stream_out

    @stream_out.setter
    def stream_out(self, value: zstreams.Sender):
        if self.stream is not None:
            raise ValueError("stream_out is set from stream,"
                             " reset stream to overwrite")
        self._stream_out = value

    @property
    def stream(self):
        """actor duplex stream, if exists"""
        return self._stream

    @stream.setter
    def stream(self, value):
        if not(isinstance(value, zstreams.Sender) and
               isinstance(value, zstreams.Receiver)):
            raise ValueError("stream must be of type Sender AND Receiver")

        if self.stream_in is not None or self.stream_out is not None:
            raise ValueError("stream_in or stream_out or both are already set")

        self.stream_in = value.stream_in
        self.stream_out = value.stream_out
        self._stream = value

    @property
    def poll_timeout(self):
        """int: actor's poller poll timeout"""
        return self._poll_timeout

    @poll_timeout.setter
    def poll_timeout(self, value):
        ensure(value).is_an(int)
        self._poll_timeout = value


class Ticker(ExecutionContext):
    """event-loop ticker"""

    @abc.abstractmethod
    def tick(self):
        """actor tick handler

        an actor tick is occurs on each event loop cycle.

        Note:
            this function should not be blocking. minimal
            behavior is recommended, if it turns out to
            be a complex function then consider breaking
            the actor into two actors.
        """

    def run(self):
        """actor process body"""
        self.initialize_context()
        self.pre_start()
        while self.runnable:
            try:
                self.tick()
            except Exception:
                self.post_stop()
                raise

        self.post_stop()


class Proxy(ExecutionContext):
    """Stream proxy actor"""

    def proxy(self):
        """The actor becomes a proxy"""
        if not(self.stream or (self.stream_in and self.stream_out)):
            raise ActorInitializationError("No streams set")

        if not self.stream:
            self.stream = zstreams.CompoundStream(
                self.stream_in, self.stream_out)

        return self.stream.run()

    def run(self):
        """actor process body"""
        self.initialize_context()
        self.pre_start()
        self.proxy()
        self.post_stop()


__all__ = [
        'ProcessEvent',
        'ExecutionContext',
        'Ticker',
        'Proxy'
        ]



"""0MQ related utilities"""


def create_socket(ctx, socket_type):
    """Creates zmq socket with default attributes
    """
    sock = ctx.socket(socket_type)
    sock.linger = 1000
    sock.rcvtimeo = 1000
    sock.sndhwm = 0
    sock.rcvhwm = 0
    return sock



"""Abstract classes/mixins for IO channels

Module contains base classes that must be derived by any channel
implementation.
"""

import abc

from motiv.sync import SystemEvent


class ChannelOutType(abc.ABC):
    """Abstract class for output channels"""

    @abc.abstractmethod
    def send(self, body, sync=True):
        "abstract send method"

    @abc.abstractmethod
    def close(self):
        "abstract close method"


class ChannelInType(abc.ABC):
    """Abstract class for input channels"""

    @abc.abstractmethod
    def receive(self):
        "abstract receive method"

    @abc.abstractmethod
    def poll(self, exit_condition: SystemEvent, poll_timeout):
        "abstract poll method"

    @abc.abstractmethod
    def close(self):
        "abstract close method"


class ChannelType(ChannelInType, ChannelOutType):
    """Abstract class for duplex channels"""


__all__ = [
        'ChannelType',
        'ChannelInType',
        'ChannelOutType'
        ]



"""Actor groups, and scalability"""

from motiv import exceptions as excs
from motiv.actor.process import ExecutionContext, ProcessEvent


class ActorGroup:
    """Wrapper for homogeneous n number of actors

    This type provides a uniform way to control a set of
    homogeneous actors
    """
    def __init__(self, name, count, actor_type, *args, **kwargs):
        if not issubclass(actor_type, ExecutionContext):
            raise TypeError(f"actor_type must be ExecutionContext,"
                            f" not {actor_type}")
        if count <= 0:
            raise ValueError("count must be a positive integer")

        self.actor_type = actor_type
        self.count = count
        self.name = name
        self.args = args
        self.kwargs = kwargs
        self.runnables = []
        self.started = ProcessEvent()
        self.stopped = ProcessEvent()

    def start(self):
        """Starts n number of actors"""
        if self.started.is_set():
            raise excs.DispatchError("group already started")

        for i in range(self.count):
            member_name = f"{self.name}[{i}]"
            actor = self.actor_type(member_name, *self.args, **self.kwargs)
            actor.start()
            self.runnables.append(actor)
        self.started.set()

    def stop(self):
        """Stops all started actors"""
        if not self.started.is_set():
            raise excs.DispatchError("group didn't started")

        if self.stopped.is_set():
            raise excs.DispatchError("group already stopped")

        for runnable in self.runnables:
            runnable.stop()
        self.stopped.set()

    def join(self):
        """Waits for all child actors to terminate"""
        for runnable in self.runnables:
            runnable.join()


class ScalableActorGroup(ActorGroup):
    """Wrapper for homogeneous dynamic actor group

    Unline `ActorGroup`, this type allows adding and removing
    a subset of child actors"""

    def scale_up(self, count):
        """adds n number of actors to current child actors"""
        if count <= 0:
            raise excs.ScalingError("count must be a positive integer")

        if self.stopped.is_set():
            raise excs.ScalingError("cannot scale a group"
                                    "after it being stopped.")

        if not self.started.is_set():
            self.count += count

        else:
            for _ in range(count):
                member_name = f"{self.name}[{self.count}]"
                self.count += 1
                actor = self.actor_type(member_name, *self.args, **self.kwargs)
                actor.start()
                self.runnables.append(actor)

    def scale_down(self, count):
        """removes/stops n number of actors from current child actors"""
        if count <= 0:
            raise excs.ScalingError("count must be positive integer")

        if count > self.count:
            raise excs.ScalingError(f"cannot scale down {count} runnables,"
                                    f" current size is {self.count}")

        if self.stopped.is_set():
            raise excs.ScalingError("cannot scale down a group"
                                    "after it being stopped.")

        if not self.started.is_set():
            self.count -= count
        else:
            stopped = []
            for _ in range(count):
                actor = self.runnables.pop()
                actor.stop()
                self.count -= 1
                stopped.append(actor)

            for actor in stopped:
                actor.join()


__all__ = [
        'ActorGroup',
        'ScalableActorGroup'
        ]



#!/usr/bin/env python3
import os

from setuptools import setup, find_packages

BASE_DIR = os.path.dirname(__file__)
REQS_PATH = os.path.join(BASE_DIR, 'requirements.txt')
README_PATH = os.path.join(BASE_DIR, "README.md")
VERSION_PATH = os.path.join(BASE_DIR, 'motiv/version.py')

with open(VERSION_PATH, "r") as fh:
    exec(fh.read())

def get_dependencies():
    install_reqs = []
    dep_links = []

    with open(REQS_PATH) as fd:
        reqs = fd.read().splitlines()

        for req in reqs:
            if req.find('+') > -1:
                dep_links.append(req)
            else:
                install_reqs.append(req)
    return install_reqs, dep_links

classifiers = [
    'Operating System :: OS Independent',
    'Natural Language :: English',
    'Intended Audience :: Developers',
    'Intended Audience :: Financial and Insurance Industry',
    'Programming Language :: Python :: 3',
]

INSTALL_REQUIRES, DEP_LINKS = get_dependencies()

setup(
    name="motiv",
    description="Motiv, Simple and Efficient processing pipelining.",
    long_description=open(README_PATH).read(),
    long_description_content_type='text/markdown',
    version=__version__,
    author="Saad Talaat",
    author_email="saadtalaat@gmail.com",
    url="https://saadtalaat.com",
    packages=find_packages(exclude=["*test*", "examples"]),
    zip_safe=False,
    install_requires= INSTALL_REQUIRES,
    dependency_links = DEP_LINKS,
    test_suite = 'nose.collector',
    platforms = 'POSIX',
    classifiers=classifiers
)




"""Channels implementation using 0MQ

Todo:
    * Check sockaddr format
"""

import os
from urllib.parse import urlparse
import zmq
from ensure import ensure_annotations, ensure

from motiv.exceptions import AlreadyConnected, NotConnected
from motiv.sync import SystemEvent, TimeoutEvent
from motiv.serde import Serializable
from motiv.channel.mixin import ChannelType, ChannelInType, ChannelOutType
from motiv.proto.zeromq import create_socket


class ChannelOut(ChannelOutType):
    """Sending only output channel

    0MQ implementation for `motiv.channel.mixin.ChannelOutType`.
    An output channel can either bind or connect to the address
    given.

    Args:
        sock_type (int): 0MQ socket_type (e.g. `zmq.PULL`)
        sockaddr (str): address to connect/bind to.

    Todo:
        * ensure sock_type is indeed a zmq socket type

    """

    @ensure_annotations
    def __init__(self, sock_type: int, sockaddr: str):
        # Internal only communication.
        url = urlparse(sockaddr)
        scheme = url.scheme
        ensure(scheme).is_in(['ipc', 'inproc', 'tcp'])
        self.address_out = sockaddr
        self.sock_connected = False
        self.sock_type = sock_type
        self.pid = os.getpid()
        self._sock_out = None

    def bind(self):
        """zmq socket binds to channel's address

        Raises:
            AlreadyConnected: channel was already connected
        """
        if self.sock_connected:
            raise AlreadyConnected("channel already initialized and connected")
        self.sock_out.bind(self.address_out)
        self.sock_connected = True

    def connect(self):
        """zmq socket connects to channel's address

        Raises:
            AlreadyConnected: channel was already connected
        """
        if self.sock_connected:
            raise AlreadyConnected("channel already initialized and connected")
        self.sock_out.connect(self.address_out)
        self.sock_connected = True

    def _send_multipart(self, body: list, sync: bool):
        flags = 0x0
        flags |= 0 if sync else zmq.NOBLOCK
        return self.sock_out.send_multipart(body, flags=flags)

    def send(self, body, sync=True):
        """sends streamed payload over the channel.

        Args:
            body (bytes, list, tuple): the payload or a list of payloads.

        Raises:
            NotConnected: channel was not connected nor binded
            TypeError: wrong body type
        """
        if not self.sock_connected:
            raise NotConnected("channel has not binded nor connected")

        result = None
        if isinstance(body, bytes):
            result = self._send_multipart([body], sync)
        elif isinstance(body, Serializable):
            payload = body.serialize()
            result = self._send_multipart([payload], sync)
        elif isinstance(body, (list, tuple)):
            frames = []
            for frame in body:
                if isinstance(frame, bytes):
                    frames.append(frame)
                elif isinstance(frame, Serializable):
                    frames.append(frame.serialize())
                else:
                    raise TypeError("Frames must be of type bytes"
                                    " or a serializable class")
            result = self._send_multipart(frames, sync)
        else:
            raise TypeError("body is not a buffer type (bytes, list, tuple)")
        return result

    def close(self):
        """closes channel"""
        self.sock_out.close()

    @property
    def sock_out(self):
        """
        Lazily evaluated
        """
        if self._sock_out is None:
            ctx = zmq.Context(2)
            self._sock_out = create_socket(ctx, self.sock_type)
        return self._sock_out


class ChannelIn(ChannelInType):
    """Receiving only input channel.

    0MQ implementation for `motiv.channel.mixin.ChannelInType`.
    An input channel can either bind or connect to the address
    given.

    Args:
        sock_type (int): 0MQ socket_type (e.g. `zmq.PULL`)
        sockaddr (str): address to connect/bind to.

    Todo:
        * ensure sock_type is indeed a zmq socket type

    """

    @ensure_annotations
    def __init__(self, sock_type: int, sockaddr: str):
        # Internal only communication.
        url = urlparse(sockaddr)
        scheme = url.scheme
        ensure(scheme).is_in(['ipc', 'inproc', 'tcp'])
        self.address_in = sockaddr
        self.pid = os.getpid()
        self.sock_connected = False
        self.sock_type = sock_type
        self._sock_in = None

    def bind(self):
        """zmq socket binds to channel's address"""
        if self.sock_connected:
            raise AlreadyConnected("channel already initialized and connected")
        self.sock_in.bind(self.address_in)
        self.sock_connected = True

    def connect(self):
        """zmq socket connects to channel's address

        Raises:
            AlreadyConnected: channel already connected.

        """
        if self.sock_connected:
            raise AlreadyConnected("channel already initialized and connected")
        self.sock_in.connect(self.address_in)
        self.sock_connected = True

    def receive(self):
        """receives data over the channel
        Note:
            this is a blocking method

        Raises:
            NotConnected: channel was not connected
        """
        if not self.sock_connected:
            raise NotConnected("channel has not binded nor connected")
        return self.sock_in.recv_multipart()

    def poll(self, poller, exit_condition: SystemEvent, timeout=-1,
             poll_timeout=50):
        """polls input socket for data received

        Note:
            this method returns as soon as it receives data or
            halt event is set.

        Args:
            poller: normally `motiv.channel.zeromq.Poller`.
            exit_condition: halt event to set to give back control to caller.
            timeout: maximum time to wait for a message.
            poll_timeout: how long to wait for messages on each iteration.

        Raises:
            NotConnected: channel was not connected.
            ValueError: wrong values (e.g. timeout < poll_timeout).
            TimeoutError: operation timed out.
        """
        if not self.sock_connected:
            raise NotConnected("channel has not binded nor connected")

        if poll_timeout > timeout >= 0:
            raise ValueError(f"timeout({timeout}) can't be"
                             " less than poll_timeout({poll_timeout})")

        ensure(exit_condition).is_a(SystemEvent)
        if self.sock_in not in poller:
            poller.register_channel(self)

        if timeout >= 0:
            return self._poll_timeout(poller, exit_condition,
                                      timeout, poll_timeout)
        return self._poll_naive(poller, exit_condition, poll_timeout)

    def _poll_timeout(self, poller, exit_condition, timeout, poll_timeout):
        timed_out = TimeoutEvent(timeout)
        timed_out.start()
        while not (exit_condition.is_set() or timed_out.is_set()):
            socks = dict(poller.poll(poll_timeout))
            if self.sock_in in socks:
                return self.receive()

        raise TimeoutError("message fetch timed out")

    def _poll_naive(self, poller, exit_condition, poll_timeout):
        while not exit_condition.is_set():
            socks = dict(poller.poll(poll_timeout))
            if self.sock_in in socks:
                return self.receive()

    def close(self):
        """closes channel"""
        self.sock_in.close()

    @property
    def sock_in(self):
        """channel's underlying input socket"""
        if not self._sock_in:
            ctx = zmq.Context(2)
            self._sock_in = create_socket(ctx, self.sock_type)
        return self._sock_in


class Channel(ChannelType):
    """Duplex channel

    0MQ implementation for `motiv.channel.mixin.ChannelType`.
    A duplex channel contains two underlying streams, input
    and output stream. these two streams can be the same stream
    or distinct streams.

    Args:
        channel_in: Input channel
        channel_out: Output channel
    """

    @ensure_annotations
    def __init__(self, channel_in: ChannelIn, channel_out: ChannelOut):
        self.cin = channel_in
        self.cout = channel_out

    def proxy(self):
        """Starts a forwarding proxy between input and output channels

        Raises:
            NotConnected: channels were not binded
        """

        if not(self.cin.sock_connected and self.cout.sock_connected):
            raise NotConnected("channels have not binded nor connected")
        ensure(self.cin.address_in).is_not_equal_to(self.cout.address_out)
        zmq.proxy(self.cin.sock_in, self.cout.sock_out)
        self.close()

    def send(self, body, sync=True):
        """Sends a payload over the output channel"""
        return self.cout.send(body, sync)

    def receive(self):
        """blocks to receive data over input channel"""
        return self.cin.receive()

    def poll(self, poller, exit_condition, poll_timeout=50):
        """polls input channel for received data"""
        return self.cin.poll(poller, exit_condition, poll_timeout=50)

    def close(self):
        """closes input, output channels"""
        self.cin.close()
        self.cout.close()


class Poller(zmq.Poller):
    """Wrapper for `zmq.Poller`"""

    @ensure_annotations
    def register_channel(self, channel: (ChannelIn, ChannelOut)):
        """registers a channel to the poller
        Args:
            channel: the channel to register.
        """
        if isinstance(channel, ChannelIn):
            self.register(channel.sock_in, zmq.POLLIN)
        if isinstance(channel, ChannelOut):
            self.register(channel.sock_out, zmq.POLLOUT)

    @ensure_annotations
    def unregister_channel(self, channel: (ChannelIn, ChannelOut)):
        """unregisters a channel from the poller

        Args:
            channel: channel to unregister
        """
        if isinstance(channel, ChannelIn):
            self.unregister(channel.sock_in)
        if isinstance(channel, ChannelOut):
            self.unregister(channel.sock_out)


__all__ = [
        'Channel',
        'ChannelIn',
        'ChannelOut',
        'Poller']



"""motiv synchronization primitives

Module:
    Using a uniform interface to define synchronization
    primitives helps us use multiple execution frameworks
    without changing any of the code written.
    for example, multiprocessing vs threading.
"""

import abc


class SystemEvent(abc.ABC):
    """Event abstract class"""

    @abc.abstractmethod
    def is_set(self):
        """checks if the event is set."""

    @abc.abstractmethod
    def set(self):
        """sets the event"""

    @abc.abstractmethod
    def clear(self):
        """clears the event"""

    @abc.abstractmethod
    def wait(self, *args, **kwargs):
        """waits till event is set"""


__all__ = [
        'SystemEvent',
        ]
